"use strict";

var _array = require("../../utils/array");

describe("moveFromTo", function () {
  var a = ["one", "two", "three"];
  it("moves elements in place", function () {
    expect((0, _array.moveFromTo)(1, 1, a)).toEqual(a);
  });
  it("moves elements forward", function () {
    expect((0, _array.moveFromTo)(0, 1, a)).toEqual(["two", "one", "three"]);
  });
  it("moves elements backward", function () {
    expect((0, _array.moveFromTo)(2, 0, a)).toEqual(["three", "one", "two"]);
  });
});
describe("modify", function () {
  it("is identity when given nothing to insert or remove", function () {
    var a = ["one", "two", "three"];
    expect((0, _array.modify)({}, [])).toEqual([]);
    expect((0, _array.modify)({}, a)).toEqual(a);
  });
  describe("inserting spans", function () {
    it("is identity when given no spans", function () {
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        insertSpans: []
      }, [])).toEqual([]);
      expect((0, _array.modify)({
        insertSpans: []
      }, a)).toEqual(a);
    });
    it("inserts a single span", function () {
      var a = ["one", "two", "three"];
      var s = [1, ["hello", "world"]];
      expect((0, _array.modify)({
        insertSpans: [s]
      }, a)).toEqual(["one", "hello", "world", "two", "three"]);
    });
    it("inserts a span at the end of the array", function () {
      var a = ["one", "two", "three"];
      var s = [3, ["the", "end"]];
      expect((0, _array.modify)({
        insertSpans: [s]
      }, a)).toEqual(["one", "two", "three", "the", "end"]);
    });
    it("errors if two spans with the same index are provided", function () {
      var redundantSpans = [[0, ["one", "thing"]], [0, ["and", "another"]]];
      expect(function () {
        (0, _array.modify)({
          insertSpans: redundantSpans
        }, []);
      }).toThrowError("at the same index");
    });
    it("inserts multiple spans in the correct place", function () {
      var s0 = [1, ["uno"]];
      var s1 = [2, ["dos"]];
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        insertSpans: [s0, s1]
      }, a)).toEqual(["one", "uno", "two", "dos", "three"]);
    });
  });
  describe("filtering", function () {
    it("is identity when given (const true)", function () {
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate() {
          return true;
        }
      }, [])).toEqual([]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate() {
          return true;
        }
      }, a)).toEqual(a);
    });
    it("removes everything when given (const false)", function () {
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate() {
          return false;
        }
      }, [])).toEqual([]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate() {
          return false;
        }
      }, a)).toEqual([]);
    });
    it("removes things based on values", function () {
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(s) {
          return s === "one";
        }
      }, [])).toEqual([]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(s) {
          return s === "one";
        }
      }, a)).toEqual(["one"]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(s) {
          return s !== "one";
        }
      }, a)).toEqual(["two", "three"]);
    });
    it("removes things based on index", function () {
      var a = ["one", "two", "three"];
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(_, i) {
          return i === 1;
        }
      }, [])).toEqual([]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(_, i) {
          return i === 1;
        }
      }, a)).toEqual(["two"]);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(_, i) {
          return i !== 1;
        }
      }, a)).toEqual(["one", "three"]);
    });
    it("removes things based on the whole array", function () {
      var a = ["one", "two", "three"];
      var pred = jest.fn();
      (0, _array.modify)({
        filterPredicate: pred
      }, a);
      expect(pred).toHaveBeenCalledWith(expect.anything(), expect.anything(), a);
      expect((0, _array.modify)({
        filterPredicate: function filterPredicate(_0, i, arr) {
          return i + 1 < arr.length && arr[i + 1] === "three";
        }
      }, a)).toEqual(["two"]);
    });
  });
  describe("inserting and filtering simultaneously", function () {
    it("can insert and remove simultaneously", function () {
      expect((0, _array.modify)({
        insertSpans: [[0, ["front"]], [3, ["middle", "content"]]],
        filterPredicate: function filterPredicate(_, i) {
          return !(i === 0 || i === 2 || i === 3);
        }
      }, ["one", "two", "three", "four", "five"])).toEqual(["front", "two", "middle", "content", "five"]);
    });
  });
});
describe("insertSpans", function () {
  it("is identity when given no spans", function () {
    var a = ["one", "two", "three"];
    expect((0, _array.insertSpans)([], [])).toEqual([]);
    expect((0, _array.insertSpans)([], a)).toEqual(a);
  });
  it("inserts a single span", function () {
    var a = ["one", "two", "three"];
    var s = [1, ["hello", "world"]];
    expect((0, _array.insertSpans)([s], a)).toEqual(["one", "hello", "world", "two", "three"]);
  });
  it("inserts a span at the end of the array", function () {
    var a = ["one", "two", "three"];
    var s = [3, ["the", "end"]];
    expect((0, _array.insertSpans)([s], a)).toEqual(["one", "two", "three", "the", "end"]);
  });
  it("errors if two spans with the same index are provided", function () {
    var redundantSpans = [[0, ["one", "thing"]], [0, ["and", "another"]]];
    expect(function () {
      (0, _array.insertSpans)(redundantSpans, []);
    }).toThrowError("at the same index");
  });
  it("inserts multiple spans in the correct place", function () {
    var s0 = [1, ["uno"]];
    var s1 = [2, ["dos"]];
    var a = ["one", "two", "three"];
    expect((0, _array.insertSpans)([s0, s1], a)).toEqual(["one", "uno", "two", "dos", "three"]);
  });
});